"""
=================
DingTalk 异常
=================
"""
from alicebot.exception import AdapterException, AdapterTimeout


class DingTalkException(AdapterException):
    """
    DingTalk 异常基类。
    """
    pass


class NetworkError(DingTalkException):
    """
    网络异常。
    """
    pass


class WebhookExpiredError(DingTalkException):
    """
    Webhook 地址已到期。
    """
    pass


class ApiTimeout(DingTalkException, AdapterTimeout):
    """
    API 请求响应超时。
    """
    pass



"""
=================
DingTalk 事件
=================
"""
import time
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, validator

from alicebot.event import Event

from .message import DingTalkMessage
from .exception import WebhookExpiredError


class UserInfo(BaseModel):
    dingtalkId: str
    staffId: Optional[str]


class Text(BaseModel):
    content: str


class DingTalkEvent(Event):
    """DingTalk 事件基类"""
    type: Optional[str] = Field(alias='msgtype')

    msgtype: str
    msgId: str
    createAt: str
    conversationType: Literal['1', '2']
    conversationId: str
    conversationTitle: Optional[str]
    senderId: str
    senderNick: str
    senderCorpId: Optional[str]
    sessionWebhook: str
    sessionWebhookExpiredTime: int
    isAdmin: Optional[bool]
    chatbotCorpId: Optional[str]
    isInAtList: Optional[bool]
    senderStaffId: Optional[str]
    chatbotUserId: str
    atUsers: List[UserInfo]
    text: Text

    message: Optional[DingTalkMessage]
    response_msg: Union[None, str, Dict, DingTalkMessage] = None
    response_at: Union[None, Dict, DingTalkMessage] = None

    @validator('message', always=True)
    def set_ts_now(cls, v, values, **kwargs):  # noqa
        return DingTalkMessage.text(values['text'].content)

    async def reply(self,
                    msg: Union[str, Dict, DingTalkMessage],
                    at: Union[None, Dict, DingTalkMessage] = None) -> Dict[str, Any]:
        """
        回复消息。

        :param msg: 回复消息的内容，可以是 str, Dict 或 DingTalkMessage。
        :param at: 回复消息时 At 的对象，必须时 at 类型的 DingTalkMessage，或者符合标准的 Dict。
        :return: 调用 Webhook 地址后钉钉服务器的响应。
        :rtype: Dict
        :exception WebhookExpiredError: 当前事件的 Webhook 地址已经过期。
        :exception ...: 同 ``DingTalkAdapter.send()`` 方法。
        """
        if self.sessionWebhookExpiredTime > time.time() * 1000:
            return await self.adapter.send(webhook=self.sessionWebhook,
                                           conversation_type=self.conversationType,
                                           msg=msg,
                                           at=at)
        else:
            raise WebhookExpiredError



"""
=================
DingTalk 消息
=================
"""
from typing import Any, Dict, List, Optional

from alicebot.message import MessageSegment


class DingTalkMessage(MessageSegment[None]):
    """DingTalk 消息"""

    @property
    def _message_class(self) -> None:
        return None

    def __str__(self):
        if self.type == 'text':
            return self.data['content']
        else:
            return super().__str__()

    def as_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        返回符合钉钉消息标准的消息字段字典。

        :return: 符合钉钉消息标准的消息字段字典。
        :rtype: Dict[str, Dict[str, Any]]
        """
        if self.type == 'raw':
            return self.data
        else:
            return {self.type: self.data}

    @classmethod
    def raw(cls, data: Dict[str, Any]) -> 'DingTalkMessage':
        """DingTalk 原始消息"""
        return cls(type='raw', data=data)

    @classmethod
    def text(cls, content: str) -> 'DingTalkMessage':
        """DingTalk text 消息"""
        return cls(type='text', data={'content': content})

    @classmethod
    def link(cls, text: str, title: str, message_url: str, pic_url: Optional[str] = None):
        """DingTalk link 消息"""
        return cls(type='link', data={'text': text,
                                      'title': title,
                                      'messageUrl': message_url,
                                      'picUrl': pic_url})

    @classmethod
    def markdown(cls, title: str, text: str):
        """DingTalk markdown 消息"""
        return cls(type='markdown', data={'title': title, 'text': text})

    @classmethod
    def action_card_single_btn(cls, title: str, text: str, single_title: str, single_url: str,
                               btn_orientation: str = '0'):
        """DingTalk 整体跳转 actionCard 消息"""
        return cls(type='actionCard', data={'title': title,
                                            'text': text,
                                            'singleTitle': single_title,
                                            'singleURL': single_url,
                                            'btnOrientation': btn_orientation})

    @classmethod
    def action_card_multi_btns(cls, title: str, text: str, btns: list, btn_orientation: str = '0'):
        """DingTalk 独立跳转 actionCard 消息"""
        return cls(type='actionCard', data={'title': title,
                                            'text': text,
                                            'btns': btns,
                                            'btnOrientation': btn_orientation})

    @classmethod
    def feed_card(cls, links: list):
        """DingTalk feedCard 消息"""
        return cls(type='feedCard', data={'links': links})

    @classmethod
    def at(cls, at_mobiles: Optional[List[str]] = None,
           at_user_ids: Optional[List[str]] = None,
           is_at_all: bool = False):
        """DingTalk At 信息"""
        return cls(type='at', data={'atMobiles': at_mobiles, 'atUserIds': at_user_ids, 'isAtAll': is_at_all})



"""
=======================
DingTalk 协议适配器
=======================
本适配器适配了钉钉企业自建机器人协议。
协议详情请参考: `钉钉开放平台`_ 。

.. _钉钉开放平台: https://developers.dingtalk.com/document/robots/robot-overview
"""
import time
import hmac
import base64
import hashlib
from typing import Any, Dict, Literal, Union

import aiohttp
from aiohttp import web

from alicebot.log import logger
from alicebot.adapter import Adapter

from .config import Config
from .event import DingTalkEvent
from .message import DingTalkMessage
from .exception import ApiTimeout, NetworkError


class DingTalkAdapter(Adapter):
    """
    钉钉协议适配器。
    """
    name: str = 'dingtalk'
    app: web.Application = None
    runner: web.AppRunner = None
    site: web.TCPSite = None

    session: aiohttp.ClientSession = None

    @property
    def config(self):
        """
        :return: 本适配器的配置。
        """
        return getattr(self.bot.config, Config.__config_name__)

    async def startup(self):
        """
        创建 aiohttp Application。
        """
        self.app = web.Application()
        self.app.add_routes([web.post(self.config.url, self.handler)])

        self.session = aiohttp.ClientSession()

    async def run(self):
        """
        运行 aiohttp 服务器。
        """
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.config.host, self.config.port)
        await self.site.start()

    async def shutdown(self):
        """
        清理 aiohttp AppRunner。
        """
        if self.session is not None:
            await self.session.close()
        if self.site is not None:
            await self.site.stop()
        if self.runner is not None:
            await self.runner.cleanup()

    async def handler(self, request: web.Request):
        """
        处理 aiohttp 服务器的接收。

        :param request: aiohttp 服务器的 Request 对象。
        """
        if 'timestamp' not in request.headers or 'sign' not in request.headers:
            logger.error(f'Illegal http header, incomplete http header')
        elif abs(int(request.headers['timestamp']) - time.time() * 1000) > 3600000:
            logger.error(f'Illegal http header, timestamp: {request.headers["timestamp"]}')
        elif request.headers['sign'] != self.get_sign(request.headers['timestamp']):
            logger.error(f'Illegal http header, sign: {request.headers["sign"]}')
        else:
            try:
                dingtalk_event = DingTalkEvent(**(await request.json()))
                dingtalk_event.adapter = self
            except Exception as e:
                logger.error(f'Request parsing error: {e!r}')
                return web.Response()
            await self.handle_event(dingtalk_event)
        return web.Response()

    def get_sign(self, timestamp: str) -> str:
        """
        计算签名。

        :param timestamp: 时间戳。
        :return: 签名。
        :rtype: str
        """
        hmac_code = hmac.new(self.config.app_secret.encode('utf-8'),
                             '{}\n{}'.format(timestamp, self.config.app_secret).encode('utf-8'),
                             digestmod=hashlib.sha256).digest()
        return base64.b64encode(hmac_code).decode('utf-8')

    async def send(self,
                   webhook: str,
                   conversation_type: Literal['1', '2'],
                   msg: Union[str, Dict, DingTalkMessage],
                   at: Union[None, Dict, DingTalkMessage] = None) -> Dict[str, Any]:
        """
        发送消息。

        :param webhook: Webhook 网址。
        :param conversation_type: 聊天类型，'1' 表示单聊，'2' 表示群聊。
        :param msg: 消息。
        :param at: At 对象，仅在群聊时生效，默认为空。
        :return: 钉钉服务器的响应。
        :rtype: Dict[str, Any]
        :exception TypeError: 传入参数类型错误。
        :exception ValueError: 传入参数值错误。
        :exception NetworkError: 调用 Webhook 地址时网络错误。
        """
        if isinstance(msg, DingTalkMessage):
            pass
        elif isinstance(msg, dict):
            msg = DingTalkMessage.raw(msg)
        elif isinstance(msg, str):
            msg = DingTalkMessage.text(msg)
        else:
            raise TypeError(f'msg must be str, Dict or DingTalkMessage, not {type(msg)!r}')

        if at is not None:
            if isinstance(at, DingTalkMessage):
                if at.type == 'at':
                    pass
                else:
                    raise ValueError(f'at.type must be "at", not {at.type}')
            elif isinstance(at, dict):
                at = DingTalkMessage.raw(at)
            else:
                raise TypeError(f'at must be Dict or DingTalkMessage, not {type(at)!r}')

        if conversation_type == '1':
            data = msg
        elif conversation_type == '2':
            if at is None:
                data = {'msgtype': msg.type, **msg.as_dict()}
            else:
                data = {'msgtype': msg.type, **msg.as_dict(), **at.as_dict()}
        else:
            raise ValueError(f'conversation_type must be "1" or "2" not {conversation_type}')

        try:
            async with self.session.post(webhook, json=data) as resp:
                return await resp.json()
        except aiohttp.ClientError:
            raise NetworkError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alicebot', 'alicebot.adapter.dingtalk']

package_data = \
{'': ['*']}

install_requires = \
['alicebot>=0.2.0,<0.3.0']

setup_kwargs = {
    'name': 'alicebot-adapter-dingtalk',
    'version': '0.1.1',
    'description': 'DingTalk adapter for AliceBot.',
    'long_description': '<div align="center">\n  <a href="https://docs.alicebot.dev/"><img src="https://raw.githubusercontent.com/st1020/alicebot/master/docs/.vuepress/public/logo.png" width="200" height="200" alt="logo"></a>\n\n# AliceBot-Adapter-DingTalk\n\n**钉钉协议适配**\n\n</div>',
    'author': 'st1020',
    'author_email': 'stone_1020@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://docs.alicebot.dev/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
